import numpy as np
import pandas as pd
from periodictable.formulas import formula

oxides = {
    "Al": "Al2O3",
    "Si": "SiO2",
    "Na": "Na2O",
    "K": "K2O",
    "Ca": "CaO",
    "Mg": "MgO",
    "Mn": "MnO",
    "Fe": "FeO",
    "Ti": "TiO2",
    "P": "P2O5",
    "Cr": "Cr2O3",
    "Y": "Y2O3",
}


def formula2wt(s):
    f = formula(s)
    s = pd.Series()
    for atom, n in f.atoms.items():
        if atom.symbol in oxides:
            ox = formula(oxides[atom.symbol])
            s[str(ox)] = n * ox.mass / ox.atoms[atom]
    return s.to_frame().T


class MineralNotCalculated(Exception):
    def __init__(self):
        super().__init__("Structural formula not yet calculated")


class Site:
    """Class to store mineral site information

    Args:
        name (str): name of site
        ncat (int): ideal number of cations
        candidates (list): list of cations in population order

    Attributes:
        name (str): name of site
        ncat (int): ideal number of cations
        candidates (list): list of cations in population order
        atoms (dict): populated atoms
        free (float): remaining space in site, i.e. `Ideal - Occupied`
    """

    def __init__(self, name, ncat, candidates):
        self.name = name
        self.ncat = ncat
        self.candidates = candidates
        self.atoms = {}

    def __repr__(self):
        return f"Site {self.name}[{self.ncat}][{sum(self.atoms.values())}]"

    def add(self, atom, amount, force=False):
        """Add atom to site

        Args:
            atom (str): cation name
            amount (float): number of attoms to add
            force (bool): If ``False`` atoms are added up to ideal occupancy. When
                ``True`` full amount is added. Default ``False``

        """
        if force:
            if atom in self.atoms:
                self.atoms[atom] += amount
            else:
                self.atoms[atom] = amount
        else:
            free = self.free
            if free > 0:
                if free > amount:
                    if atom in self.atoms:
                        self.atoms[atom] += amount
                    else:
                        self.atoms[atom] = amount
                else:
                    self.atoms[atom] = free

    def get(self, atom, fraction=False):
        """Get number of given atom on site

        Args:
            atom (str): atom name

        """
        if fraction:
            return self.atoms.get(atom, 0) / self.ncat
        else:
            return self.atoms.get(atom, 0)

    @property
    def free(self):
        return self.ncat - sum(self.atoms.values())


class StrucForm:
    """Class to manipulate mineral structural formula

    Args:
        mineral (Mineral): mineral instance

    Attributes:
        mineral (Mineral): mineral instance
        sites (list): list of sites
        reminder (pandas.Series): remainder after site population or empty
    """

    def __init__(self, mineral):
        self.mineral = mineral
        self.sites = [Site(*s) for s in self.mineral.structure]
        self.reminder = pd.Series()

    def __repr__(self):
        if not self.reminder.empty:
            res = ""
            for site in sorted(self.sites, key=lambda site: site.name):
                s = ""
                for atom, val in site.atoms.items():
                    s += f"{atom}({val:.3f})"
                res += f"[{s}]"
            return res
        else:
            return "Not yet calculated"

    def get(self, atom) -> np.floating:
        """Get total number of given atom in structure

        Args:
            atom (str): atom name

        """
        return np.sum([s.get(atom) for s in self.sites])

    def site(self, name) -> Site | None:
        """Get site by name

        Args:
            name (str): site name

        """
        sites = [s for s in self.sites if s.name == name]
        if sites:
            return sites[0]

    @property
    def apfu(self) -> pd.Series:
        """Calculate mineral atom p.f.u based on site occupancies

        Returns:
            pandas.Series: atoms p.f.u
        """
        if not self.reminder.empty:
            apfu = pd.Series(
                [self.get(e) for e in self.reminder.index], index=self.reminder.index
            )
            # Add atoms not in analysis
            for site in self.sites:
                for atom in site.candidates:
                    if atom not in apfu:
                        apfu[atom] = 0.0
            return apfu
        else:
            raise MineralNotCalculated()

    @property
    def check_stechiometry(self) -> np.floating:
        """Calculate average missfit of populated and ideal cations on sites"""
        if not self.reminder.empty:
            return np.mean([abs(s.free) / s.ncat for s in self.sites])
        else:
            raise MineralNotCalculated()


class Mineral:
    """Base Mineral class to store mineral structure and endmembers calculation

    Attributes:
        noxy (int): number of oxygens
        ncat (int): number of cations
        needsFe (str): `"Fe2"` for only FeO oxide needed, `"Fe3"` for both FeO and
            Fe2O3 oxide needed, `None` for no Fe needed
        structure (tuple, optional): structural formula definition. Iterable collection
            of tuples `(name, ncat, list of cations)` e.g.
            `("T", 4, ["Si{4+}", "Al{3+}"])`. Tuples are ordered in population order.
        has_structure (bool): True when structure is defined
        has_endmembers (bool): True when endmembers method is defined

    """

    def __init__(self):
        # Subclasses should define this!
        self.structure = ()
        self.noxy = 0
        self.needsFe = False

    def __repr__(self):
        return type(self).__name__

    @property
    def ncat(self):
        return sum([s[1] for s in self.structure])

    @property
    def has_endmembers(self):
        return hasattr(self, "endmembers")

    @property
    def has_structure(self):
        return self.structure != ()

    def endmembers(self, cations, force=False):
        raise NotImplementedError("Subclasses should implement this!")

    def calculate(self, cations, force=False) -> StrucForm:
        """Calculate structural formula from cations p.f.u

        Args:
            cations (pandas.Series): cations p.f.u

        Returns:
            StrucForm: structural formula

        """
        sf = StrucForm(self)
        lastsite = {}
        for site in sf.sites:
            for atom in site.candidates:
                available = cations.get(atom, 0.0) - sf.get(atom)
                if available > 0:
                    site.add(atom, available)
                    lastsite[atom] = site

        # Force site occupancy
        if force:
            occ = pd.Series([sf.get(e) for e in cations.index], index=cations.index)
            reminder = cations - occ
            for atom, r in reminder.items():
                if (r > 0) & (atom in lastsite):
                    lastsite[atom].add(atom, r, force=True)
        # reminder
        occ = pd.Series([sf.get(e) for e in cations.index], index=cations.index)
        sf.reminder = cations - occ
        return sf

    def apfu(self, cations, force=False) -> pd.Series:
        """Calculate mineral atom p.f.u based on site occupancies

        Args:
            cations (pandas.Series): cations p.f.u

        Returns:
            pandas.Series: atoms p.f.u
        """
        sf = self.calculate(cations, force=force)
        return sf.apfu

    def check_stechiometry(self, cations, force=False) -> np.floating:
        """Calculate average missfit of populated and ideal cations on sites

        Args:
            cations (pandas.Series): cations p.f.u

        Returns:
            pandas.Series: average missfit
        """
        sf = self.calculate(cations, force=force)
        return sf.check_stechiometry


class Garnet_Fe2(Mineral):
    """Garnet using total Fe with 4 endmembers"""

    def __init__(self):
        super().__init__()
        self.noxy = 12
        self.needsFe = "Fe2"
        # fmt: off
        self.structure = (
            ("Z", 3, ["Si{4+}", "Al{3+}"]),
            ("Y", 2, ["Si{4+}", "Al{3+}", "Ti{4+}", "Cr{3+}", "Mg{2+}", "Fe{2+}", "Mn{2+}"]),
            ("X", 3, ["Y{3+}", "Mg{2+}", "Fe{2+}", "Mn{2+}", "Ca{2+}", "Na{+}"]),
        )
        # fmt: on

    def endmembers(self, cations, force=False):
        apfu = self.apfu(cations, force=force)
        esum = apfu["Fe{2+}"] + apfu["Mn{2+}"] + apfu["Mg{2+}"] + apfu["Ca{2+}"]
        em = {
            "Alm": apfu["Fe{2+}"] / esum,
            "Prp": apfu["Mg{2+}"] / esum,
            "Sps": apfu["Mn{2+}"] / esum,
            "Grs": apfu["Ca{2+}"] / esum,
        }
        return pd.Series(em)


class Garnet(Mineral):
    """Garnet using Fe2 and Fe3 with 6 endmembers"""

    def __init__(self):
        super().__init__()
        self.noxy = 12
        self.needsFe = "Fe3"
        # fmt: off
        self.structure = (
            ("Z", 3, ["Si{4+}", "Al{3+}", "Fe{3+}"]),
            ("Y", 2, ["Si{4+}", "Al{3+}", "Ti{4+}", "Cr{3+}", "Fe{3+}", "Mg{2+}", "Fe{2+}", "Mn{2+}"]),
            ("X", 3, ["Y{3+}", "Mg{2+}", "Fe{2+}", "Mn{2+}", "Ca{2+}", "Na{+}"]),
        )
        # fmt: on

    def endmembers(self, cations, force=False):
        apfu = self.apfu(cations, force=force)
        s8t = apfu["Fe{2+}"] + apfu["Mn{2+}"] + apfu["Mg{2+}"] + apfu["Ca{2+}"]
        s6t = apfu["Ti{4+}"] + apfu["Al{3+}"] + apfu["Cr{3+}"] + apfu["Fe{3+}"]
        # end members in mol%
        em = {
            "Alm": apfu["Fe{2+}"] / s8t,
            "Prp": apfu["Mg{2+}"] / s8t,
            "Sps": apfu["Mn{2+}"] / s8t,
            "Grs": (apfu["Al{3+}"] / s6t) * (apfu["Ca{2+}"] / s8t),
            "Adr": (apfu["Fe{3+}"] / s6t) * (apfu["Ca{2+}"] / s8t),
            "Uv": (apfu["Cr{3+}"] / s6t) * (apfu["Ca{2+}"] / s8t),
            "CaTi": (apfu["Ti{4+}"] / s6t) * (apfu["Ca{2+}"] / s8t),
        }
        return pd.Series(em)


class Feldspar(Mineral):
    """Feldspar with 3 endmembers"""

    def __init__(self):
        super().__init__()
        self.noxy = 8
        self.needsFe = None
        self.structure = (
            ("T", 4, ["Si{4+}", "Al{3+}"]),
            ("A", 1, ["K{+}", "Na{+}", "Ca{2+}"]),
        )

    def endmembers(self, cations, force=False):
        apfu = self.apfu(cations, force=force)
        alk = apfu["Ca{2+}"] + apfu["Na{+}"] + apfu["K{+}"]
        em = {
            "An": apfu["Ca{2+}"] / alk,
            "Ab": apfu["Na{+}"] / alk,
            "Or": apfu["K{+}"] / alk,
        }
        return pd.Series(em)


class Pyroxene_Fe2(Mineral):
    """Simple Pyroxene using total Fe with 3 endmembers"""

    def __init__(self):
        super().__init__()
        self.noxy = 6
        self.needsFe = "Fe2"
        self.structure = (
            ("T", 2, ["Si{4+}", "Al{3+}"]),
            ("M1", 1, ["Al{3+}", "Ti{4+}", "Cr{3+}", "Mn{2+}", "Mg{2+}", "Fe{2+}"]),
            ("M2", 1, ["Mg{2+}", "Fe{2+}", "Ca{2+}", "Na{+}", "K{+}"]),
        )

    def endmembers(self, cations, force=False):
        apfu = self.apfu(cations, force=force)
        esum = apfu["Fe{2+}"] + apfu["Mn{2+}"] + apfu["Mg{2+}"] + apfu["Ca{2+}"]
        em = {
            "En": apfu["Mg{2+}"] / esum,
            "Wo": apfu["Ca{2+}"] / esum,
            "Fs": (apfu["Fe{2+}"] + apfu["Mn{2+}"]) / esum,
        }
        return pd.Series(em)


class Pyroxene(Mineral):
    """Pyroxene"""

    def __init__(self):
        super().__init__()
        self.noxy = 6
        self.needsFe = "Fe3"
        # fmt: off
        self.structure = (
            ("T", 2, ["Si{4+}", "Al{3+}", "Fe{3+}"]),
            ("M1", 1, ["Al{3+}", "Ti{4+}", "Fe{3+}", "Cr{3+}", "Mn{2+}", "Mg{2+}", "Fe{2+}"]),
            ("M2", 1, ["Mg{2+}", "Fe{2+}", "Ca{2+}", "Na{+}", "K{+}"]),
        )
        # fmt: on

    def endmembers(self, cations, force=False):
        apfu = self.apfu(cations, force=force)
        acm = min(apfu["Na{+}"] + apfu["K{+}"], apfu["Fe{3+}"])
        nar = apfu["Na{+}"] + apfu["K{+}"] - acm
        al4 = np.clip(2 - apfu["Si{4+}"], 0, apfu["Al{3+}"])
        al6 = apfu["Al{3+}"] - al4
        jad = min(al6, nar)
        cati = min(apfu["Ti{4+}"], al4 / 2)
        al4r = al4 - 2 * cati
        al6r = al6 - 2 * cati
        cacr = min(apfu["Cr{3+}"], al4r)
        fe3r = apfu["Fe{3+}"] - acm
        cafe = min(al4r, fe3r)
        cats = min(al4r, al6r)
        car = apfu["Ca{2+}"] - (cati + cacr + cafe + cats)
        em = {
            "Aeg": acm,
            "Jd": jad,
            "CaTi": cati,
            "CaCr": cacr,
            "Hd": cafe,
            "CaTs": cats,
            "Esc": 0.5 * car if al6r > (2 * car) else al6r,
            "En": apfu["Mg{2+}"] / 2,
            "Fs": apfu["Fe{2+}"] / 2,
            "CaMn": apfu["Mn{2+}"] / 2,
            "Wo": car / 2 if car > 0 else 0,
        }
        return pd.Series(em)


class DiMica(Mineral):
    """DiOct Micas"""

    def __init__(self):
        super().__init__()
        self.noxy = 11
        self.needsFe = "Fe2"
        self.structure = (
            ("T", 4, ["Si{4+}", "Al{3+}"]),
            ("M", 2, ["Al{3+}", "Ti{4+}", "Cr{3+}", "Fe{2+}", "Mn{2+}", "Mg{2+}"]),
            ("I", 1, ["Ba{2+}", "Ca{2+}", "Na{+}", "K{+}"]),
        )

    def endmembers(self, cations, force=False):
        sf = self.calculate(cations, force=force)
        apfu = self.apfu(cations, force=force)
        Xm = sf.site("M").get("Al{3+}") - 1
        XCel = 1 - Xm  # total amout (Fe Al-Celadonite + Mg Al-Celadonite)
        XMg = apfu["Mg{2+}"] / (apfu["Mg{2+}"] + apfu["Fe{2+}"])
        Isum = apfu["Ca{2+}"] + apfu["Na{+}"] + apfu["K{+}"]
        XMPM = (
            Isum * Xm
        )  # Sum of Ca + Na + K multiplied by the total fraction of Ms, Pg, Mrg, and Prl
        em = {
            "Al-Celadonite": XMg * XCel,
            "Fe Al-Celadonite": XCel - XMg * XCel,
            "Pyrophyllite": Xm - XMPM,
            "Margarite": XMPM * apfu["Ca{2+}"] / Isum,
            "Paragonite": XMPM * apfu["Na{+}"] / Isum,
            "Muscovite": XMPM * apfu["K{+}"] / Isum,
        }
        return pd.Series(em)



import importlib.resources
import json

import matplotlib.pyplot as plt
import mpltern  # NOQA
import numpy as np
import pandas as pd
import pyparsing
import seaborn as sns
from matplotlib.ticker import MaxNLocator
from pandas.api.types import is_numeric_dtype
from periodictable import elements
from periodictable.core import iselement, ision
from periodictable.formulas import formula

from petropandas.constants import (
    AGECOLS,
    COLNAMES,
    ISOPLOT,
    ISOPLOT_FORMATS,
    REE,
    REE_PLOT,
)
from petropandas.minerals import Mineral

config = {
    "isoplot_default_format": 2,
    "colnames": COLNAMES,
    "agecols": AGECOLS,
    "isoplot_formats": ISOPLOT_FORMATS,
    "ree_plot": REE_PLOT,
    "sort_oxides": False,
    "oxides_order": [
        "SiO2",
        "Al2O3",
        "FeO",
        "Fe2O3",
        "CaO",
        "Na2O",
        "MgO",
        "K2O",
        "TiO2",
        "P2O5",
        "MnO",
        "BaO",
        "ZrO2",
        "VO2",
        "Cr2O3",
        "ZnO",
        "Y2O3",
        "Sc2O3",
        "PbO",
    ],
}

germ = importlib.resources.files("petropandas").joinpath("data", "germ.json")
with germ.open() as fp:
    config["reservoirs"] = json.load(fp)

ionox = {
    1: ("2", ""),
    2: ("", ""),
    3: ("2", "3"),
    4: ("", "2"),
    5: ("2", "5"),
    6: ("", "3"),
    7: ("2", "7"),
    8: ("", "4"),
}


def oxideprops(f):
    ncat, element = f.structure[0]
    noxy = f.atoms[elements.name("oxygen")]
    charge = 2 * noxy // ncat
    return {
        "mass": f.mass,
        "cation": element.ion[charge],
        "noxy": noxy,
        "ncat": ncat,
        "charge": charge,
        "elfrac": f.mass_fraction[element],
    }


def elementprops(f):
    return {
        "mass": f.mass,
        "charge": f.charge,
    }


class MissingColumns(Exception):
    def __init__(self, col):
        super().__init__(f"Must have {col} in columns.")


class NotTextualIndex(Exception):
    def __init__(self):
        super().__init__("Index is not textual.")


class NotTextualColumn(Exception):
    def __init__(self, col):
        super().__init__(f"Column {col} is not textual.")


class TemplateNotDefined(Exception):
    def __init__(self, tmpl):
        super().__init__(
            f"Column definition {tmpl} is not defined. Check `config['colnames']`"
        )


class NoEndMembers(Exception):
    def __init__(self, mineral):
        super().__init__(f"Mineral {mineral} has no endmembers method defined.")


@pd.api.extensions.register_dataframe_accessor("petro")
class PetroAccessor:
    """Use `.petro` pandas dataframe accessor."""

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def search(self, s, **kwargs) -> pd.DataFrame:
        """Select subset of data from dataframe containing string s in index or column.

        Note: Works only with non-numeric index or column

        Args:
            s (str): Returns all rows which contain string s in index or column.

        Keyword Args:
            on (str): Name of column used for search. When `None` the index is used
            regex (bool): If True, assumes the pat is a regular expression. If False,
                treats the pat as a literal string.

        Returns:
            Dataframe with selected data
        """
        on = kwargs.get("on", None)
        regex = kwargs.get("regex", True)
        if on is None:
            col = self._obj.index
        else:
            if on not in self._obj:
                raise MissingColumns(on)
            col = self._obj[on]
        if not is_numeric_dtype(col):
            col = pd.Series([str(v) for v in col], index=self._obj.index)
            return self._obj.loc[col.str.contains(s, regex=regex).fillna(False)].copy()
        else:
            if on is None:
                raise NotTextualIndex()
            else:
                raise NotTextualColumn(on)

    def fix_columns(self, template) -> pd.DataFrame:
        """Rename columns according to predefined template.

        Check `config['colnames']` for available templates. User-defined templates
        could be added. Template is a dict used for `pandas.DataFrame.rename`.

        Args:
            template (str): Name of renaming template

        Returns:
            Dataframe with renamed columns
        """
        if template not in config["colnames"]:
            raise TemplateNotDefined(template)
        return self._obj.rename(columns=config["colnames"][template])

    def strip_columns(self) -> pd.DataFrame:
        """Strip whitespaces from column names

        Returns:
            Dataframe with stripped column names
        """
        return self._obj.rename(columns=lambda x: x.strip())

    def calc(self, expr, name=None) -> pd.DataFrame:
        """Calculate a new column using expression

        Evaluate a string describing operations on DataFrame columns.

        Args:
            expr (str): The expression string to evaluate.
            name (str): Name of column to store result. When `None` the expression
                is used as name.

        Returns:
            Dataframe with calculated column
        """
        self._obj[expr if name is None else name] = self._obj.eval(expr)
        return self._obj

    def to_latex(self, total=False, transpose=True, precision=2) -> str:
        """Convert datatable to LaTeX string

        Args:
            total (bool, optional): Add column `"Total"` with total sums.
                Default `True`
            transpose (bool, optional): Place samples as columns. Default ``True``
            precision (bool, optional): Nimber of decimal places. Default 2

        """
        df = self._obj.copy()
        if total:
            df["Total"] = df.sum(axis=1, numeric_only=True)
        if transpose:
            df = df.T
        return df.fillna("").style.format(precision=precision).to_latex()


@pd.api.extensions.register_dataframe_accessor("petroplots")
class PetroPlotsAccessor:
    """Use `.petroplots` pandas dataframe accessor."""

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def profile(self, **kwargs):
        """Plot garnet profiles.

        Note:
            DataFrame must contains garnet endmembers properly sorted.

        Keyword Args:
            use_index (bool): When True, xticks are derived from DataFrame
                index, otherwise ticks are sequential. Default False
            extra (list): list of columns on secondary axis.
                Default `[]`
            lim (tuple): y-axis limits for principal axes or auto when
                ``None``. Default ``None``
            lim_extra (tuple): y-axis limits for secondary axes or auto when
                ``None``. Default ``None``
            percents (bool): When ``True`` y-axes scale is percents, otherwise fraction
            xlabel (str): label of the x-axis. Default auto.
            filename (str): When not ``None``, the plot is saved to file,
                otherwise the plot is shown.
            maxticks (int): maximum number of ticks on x-axis. Default 20
            xticks_rotation (int): rotation of xticks labels. Default 0
            markers (list): Markers used. Default None
            subplot_kws (dict): kwargs passed to matplotlib subplots. Default {}
            grid (bool): Show grid. Default False
            grid_kws (dict): kwargs passed to matplotlib grid.
                Default dict(visible=True)
            grid_ticks (int): number of ticks on twin axes, when grid is True.
                Default 10

        """
        em = self._obj.copy()

        cols = em.columns.tolist()
        extra = kwargs.get("extra", [])
        margin = kwargs.get("margin", 0.1)
        lim = kwargs.get("lim", None)
        lim_extra = kwargs.get("lim_extra", None)
        filename = kwargs.get("filename", None)
        maxticks = kwargs.get("maxticks", 20)
        percents = kwargs.get("percents", False)
        use_index = kwargs.get("use_index", False)
        xlabel = kwargs.get("xlabel", None)
        xticks_rotation = kwargs.get("xticks_rotation", 0)
        markers = kwargs.get("markers", None)

        # validate
        cols_extra = []
        for c in extra:
            if c in cols:
                cols.remove(c)
                cols_extra.append(c)
        twin = True if cols_extra else False

        colors = sns.color_palette(None, len(cols) + len(cols_extra))
        fig, ax1 = plt.subplots(**kwargs.get("subplot_kws", {}))
        if percents:
            multiple = 100
            unit = " [%]"
        else:
            multiple = 1
            unit = " [fraction]"
        if use_index:
            xlabel = "index" if xlabel is None else xlabel
        else:
            em.index = range(len(em))
            xlabel = "position" if xlabel is None else xlabel
        ax1.set_xlabel(xlabel)
        if markers is None:
            markers1 = markers2 = None
        else:
            if len(markers) == 1:
                markers = (len(cols) + len(cols_extra)) * markers
            markers1 = markers[: len(cols)]
            markers2 = markers[len(cols) :]
        if twin:
            ax1.set_ylabel(" ".join(cols) + unit)
            with sns.color_palette(colors[: len(cols)]):
                lns1 = sns.lineplot(
                    multiple * em[cols], ax=ax1, markers=markers1, dashes=False
                )
            # h1 = ax1.plot(xvals, multiple * em[cols], marker=marker, ms=ms)
            if lim is not None:
                ax1.set_ylim(lim[0], lim[1])
            else:
                ax1.set_ymargin(margin)
            ax2 = ax1.twinx()
            ax2.set_ylabel(" ".join(cols_extra) + unit)
            with sns.color_palette(colors[len(cols) :]):
                lns2 = sns.lineplot(
                    multiple * em[cols_extra],
                    ax=ax2,
                    markers=markers2,
                    dashes=False,
                )
            # h2 = ax2.plot(xvals, multiple * em[cols_extra], marker=marker, ms=ms)
            if lim_extra is not None:
                ax2.set_ylim(lim_extra[0], lim_extra[1])
            else:
                ax2.set_ymargin(margin)
            # common legend
            h1, l1 = lns1.get_legend_handles_labels()
            h2, l2 = lns2.get_legend_handles_labels()
            ax1.legend(
                handles=h1 + h2,
                labels=l1 + l2,
                loc=8,
                bbox_to_anchor=(0.5, 1),
                ncol=4,
                title=None,
                frameon=False,
            )
            ax2.get_legend().remove()
        else:
            ax1.set_ylabel(unit[2:-1])
            with sns.color_palette(colors):
                lns1 = sns.lineplot(
                    multiple * em[cols + cols_extra],
                    ax=ax1,
                    markers=markers,
                    dashes=False,
                )
            # h1 = ax1.plot(xvals, multiple * em[cols + cols_extra], marker=marker, ms=ms)
            if lim is not None:
                ax1.set_ylim(lim[0], lim[1])
            else:
                ax1.set_ymargin(margin)
            sns.move_legend(
                ax1, loc=8, bbox_to_anchor=(0.5, 1), ncol=4, title=None, frameon=False
            )
        # Find at most maxticks ticks on the x-axis at 'nice' locations
        xloc = MaxNLocator(maxticks - 1, integer=True)
        ax1.xaxis.set_major_locator(xloc)
        ax1.tick_params(axis="x", labelrotation=xticks_rotation)
        # grid
        if kwargs.get("grid", False):
            if twin:
                nticks = kwargs.get("grid_ticks", 10)
                ax1.set_yticks(
                    np.linspace(ax1.get_ybound()[0], ax1.get_ybound()[1], nticks)
                )
                ax2.set_yticks(
                    np.linspace(ax2.get_ybound()[0], ax2.get_ybound()[1], nticks)
                )
            ax1.grid(kwargs.get("grid_kws", dict(visible=True)))
        fig.tight_layout()
        if filename is not None:
            fig.savefig(filename)
            print(f"{filename} saved.")
            plt.close(fig)
        else:
            plt.show()

    def ternary(self, *args, **kwargs):
        """Ternary scatter plot

        Args:
            top (str): name of column or expression for top variable.
                Default column 0
            left (str): name of column or expression for left variable.
                Default column 1
            right (str): name of column or expression for right variable.
                Default column 2

        Keyword Args:
            kind (str): kind of plot, `scatter`, `plot` or `contour`. Other
                values returns empty plot. Deafult `scatter`.
            c (str | array like): values used for coloring as expression (str)
                or array of values.
            s (str | array like): values used for sizing as expression (str)
                or array of values.
            v (str | array like): values used for contouring as expression (str)
                or array of values.
            ternary_sum (float): Total sum. Default 1.0
            tlim (tuple): top limits. Default(0, 1)
            llim (tuple): top limits. Default(0, 1)
            rlim (tuple): top limits. Default(0, 1)
            grid (bool): show grid. Default False
            figsize (tuple): width, height in inches. If not provided, defaults to
                rcParams["figure.figsize"]
            ax (Axes): matplotlib axes to be used.
            return_ax (bool): Whether to return matplotlib axes.
                Default False
            show (bool): Whether to show plot. Default True

        Additional keyword arguments are passed to underlying matplotlib
        function, e.g. `s` or `c` to `scatter` or `label`

        """
        if len(args) > 0:
            top = args[0]
        else:
            top = self._obj.columns[0]
        if len(args) > 1:
            left = args[1]
        else:
            left = self._obj.columns[1]
        if len(args) > 2:
            right = args[2]
        else:
            right = self._obj.columns[2]
        kind = kwargs.pop("kind", "scatter")
        ternary_sum = kwargs.pop("ternary_sum", 1.0)
        return_ax = kwargs.pop("return_ax", False)
        show = kwargs.pop("show", True)
        filename = kwargs.pop("filename", None)
        figsize = kwargs.pop("figsize", plt.rcParams["figure.figsize"])
        grid = kwargs.pop("grid", False)
        if "ax" in kwargs:
            ax = kwargs.pop("ax")
        else:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(projection="ternary", ternary_sum=ternary_sum)
        ax.set_tlim(*kwargs.pop("tlim", (0, ternary_sum)))
        ax.set_llim(*kwargs.pop("llim", (0, ternary_sum)))
        ax.set_rlim(*kwargs.pop("rlim", (0, ternary_sum)))
        ax.set_tlabel(top)
        ax.set_llabel(left)
        ax.set_rlabel(right)
        top_vals = self._obj.eval(top)
        left_vals = self._obj.eval(left)
        right_vals = self._obj.eval(right)
        leg_color = False
        leg_size = False
        tit_leg = None
        match kind:
            case "scatter":
                if "s" in kwargs:
                    if isinstance(kwargs["s"], str):
                        tit_leg = kwargs["s"]
                        kwargs["s"] = self._obj.eval(kwargs["s"])
                if "c" in kwargs:
                    if isinstance(kwargs["c"], str):
                        tit_leg = kwargs["c"]
                        kwargs["c"] = self._obj.eval(kwargs["c"])
                pc = ax.scatter(top_vals, left_vals, right_vals, **kwargs)
                if "c" in kwargs:
                    handles, labels = pc.legend_elements(prop="colors", num=6)
                    if len(handles) > 1:
                        leg_color = ax.legend(
                            handles, labels, loc="upper right", title=tit_leg
                        )
                if "s" in kwargs:
                    handles, labels = pc.legend_elements(prop="sizes", num=6)
                    if len(handles) > 1:
                        leg_size = ax.legend(
                            handles, labels, loc="upper left", title=tit_leg
                        )
            case "plot":
                pc = ax.plot(top_vals, left_vals, right_vals, **kwargs)
            case "contourf":
                if "v" in kwargs:
                    if isinstance(kwargs["v"], str):
                        tit_leg = kwargs["c"]
                        v = self._obj.eval(kwargs["v"])
                    else:
                        v = kwargs["v"]
                    pc = ax.tricontourf(top_vals, left_vals, right_vals, v, **kwargs)
                    cax = ax.inset_axes([1.05, 0.1, 0.05, 0.9], transform=ax.transAxes)
                    colorbar = fig.colorbar(pc, cax=cax)
                    colorbar.set_label(tit_leg, rotation=270, va="baseline")
        if grid:
            ax.grid()
        if return_ax:
            return ax
        if show:
            handles, labels = ax.get_legend_handles_labels()
            if handles:
                ax.legend(handles, labels)
            else:
                if leg_color:
                    ax.add_artist(leg_color)
                if leg_size:
                    ax.add_artist(leg_size)
            if filename is not None:
                ax.get_figure().savefig(filename)
                print(f"{filename} saved.")
                plt.close(ax.get_figure())
            else:
                plt.show()


@pd.api.extensions.register_dataframe_accessor("isoplot")
class IsoplotAccessor:
    """Use `.isoplot` pandas dataframe accessor."""

    def __init__(self, pandas_obj):
        self._names = []
        self._others = []
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        # verify there is a isoplot column
        valid = []
        for col in obj.columns:
            if col in ISOPLOT:
                valid.append(True)
                self._names.append(col)
            else:
                valid.append(False)
                self._others.append(col)
        if not any(valid):
            raise MissingColumns("isoplot")

    def df(self, **kwargs) -> pd.DataFrame:
        """Returns dataframe.

        Keyword Args:
            iso (int): IsoplotR format. Default `config["isoplot_default_format"]`
            C (str): Column to be used as color. Default None
            omit (str): Column to be used as omit. Default None
            comment (str): Column to be used as comment. Default None

        Returns:
            Dataframe
        """
        iso = kwargs.get("iso", config["isoplot_default_format"])
        df = self._obj[config["isoplot_formats"][iso]]
        if "C" in kwargs:
            df["C"] = self._obj[kwargs["C"]]
        else:
            df["C"] = None
        if "omit" in kwargs:
            df["omit"] = self._obj[kwargs["omit"]]
        else:
            df["omit"] = None
        if "comment" in kwargs:
            df["comment"] = self._obj[kwargs["comment"]]
        else:
            df["comment"] = None
        return df

    def clipboard(self, **kwargs):
        """Copy data to clipbord to be used in IsoplotR online.

        Note:
            [IsoplotR online](http://isoplotr.es.ucl.ac.uk/home/index.html)

                Vermeesch, P., 2018, IsoplotR: a free and open toolbox for geochronology.
                Geoscience Frontiers, v.9, p.1479-1493, doi: 10.1016/j.gsf.2018.04.001.

        Keyword Args:
            iso (int): IsoplotR format. Default `config["isoplot_default_format"]`
            C (str): Column to be used as color. Default None
            omit (str): Column to be used as omit. Default None
            comment (str): Column to be used as comment. Default None
        """
        df = self.df(**kwargs)
        df.to_clipboard(header=False, index=False)

    def calc_ages(self, **kwargs) -> pd.DataFrame | None:
        """Copy data to clipbord, calc ages in IsoplotR online and paste back results.

        Keyword Args:
            iso (int): IsoplotR format. Default `config["isoplot_default_format"]`
            C (str): Column to be used as color. Default None
            omit (str): Column to be used as omit. Default None
            comment (str): Column to be used as comment. Default None

        Returns:
            Dataframe with calculated ages
        """
        iso = kwargs.get("iso", config["isoplot_default_format"])
        self.clipboard(**kwargs)
        print(f"Data in format {iso} copied to clipboard")
        print("Calc ages with Stacey-Kramers, discordance and digits 5")
        input("Then copy to clipboard and press Enter to continue...")
        ages = pd.read_clipboard(header=None)
        if ages.shape[1] == 9:
            ages.columns = config["agecols"]
            ages.index = self._obj.index
            for col in config["agecols"]:
                self._obj[col] = ages[col]
                self._validate(self._obj)
            print("Ages added to data")
        else:
            print(
                f"Wrong shape {ages.shape} of copied data. Awaits {self._obj.shape} Set correct options and try again."
            )


class AccessorTemplate:
    """Base class"""

    def __init__(self, pandas_obj):
        self._names = []
        self._others = []
        self._names_props = []
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @property
    def _df(self) -> pd.DataFrame:
        """Returns dataframe with appropriate columns."""
        return self._obj[self._names]

    def _validate(self, obj):
        raise NotImplementedError

    def _final(self, df, **kwargs) -> pd.DataFrame:
        # select = kwargs.get("select", [])
        # if select:
        #     df = df[df.columns.intersection(select)]
        #     rest = df.columns.symmetric_difference(select).difference(df.columns)
        #     df[rest] = np.nan
        keep = kwargs.get("keep", [])
        dropna = kwargs.get("dropna", True)
        match keep:
            case []:
                pass
            case "all":
                df = pd.concat([df, self._obj[self._others]], axis=1)
            case _:
                df = pd.concat([df, self._obj[keep]], axis=1)
        if dropna:
            return df.dropna(axis=1, how="all")
        else:
            return df

    @property
    def props(self) -> pd.DataFrame:
        """Returns properties."""
        return pd.DataFrame(self._names_props, index=pd.Index(self._names))

    def df(self, **kwargs) -> pd.DataFrame:
        """Returns dataframe.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe
        """
        return self._final(self._df, **kwargs)

    def dropna(self, **kwargs) -> pd.DataFrame:
        """Drop columns with NA values.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].

        Returns:
            Dataframe without NA columns
        """
        return self._final(self._df.dropna(axis=1), **kwargs)

    def molprop(self, **kwargs) -> pd.DataFrame:
        """Convert oxides weight percents to molar proportions.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with molar proportions

        """
        res = self._df.div(self.props["mass"])
        return self._final(res, **kwargs)

    def mean(self) -> pd.DataFrame:
        """Return Dataframe with single row of arithmetic means of valid columns"""
        return (self._df.sum(axis=0) / len(self._df)).to_frame().T

    def sum(self) -> pd.DataFrame:
        """Return Dataframe with single row of sum of valid columns"""
        return self._df.sum(axis=0).to_frame().T

    def scale(self, **kwargs) -> pd.DataFrame:
        """Normalize values to given sum.

        Keyword Args:
            to (float): Sum of values. Default 100.0
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Scaled dataframe
        """
        to = kwargs.get("to", 100.0)
        res = to * self._df.div(self._df.sum(axis=1), axis=0)
        return self._final(res, **kwargs)

    def plot(self, **kwargs):
        """Paiplot of data.

        Notes: All keywords except further listed are passed to seaborn pairplot.
            Use all accessor valid columns or columns provided in `vars`

        Keyword Args:
            title (str): Title of the plot. Default None
            filename (str): If not none, plot is saved to file. Default None.
            dpi (int): DPI used for `savefig`. Default 150.
        """
        title = kwargs.pop("title", None)
        filename = kwargs.pop("filename", None)
        dpi = kwargs.pop("filename", 150)

        if "vars" not in kwargs:
            kwargs["vars"] = self._names
            keep = []
        else:
            keep = [c for c in kwargs["vars"] if c not in self._names]

        if "hue" in kwargs:
            keep.append(kwargs["hue"])

        g = sns.pairplot(self.df(keep=keep), **kwargs)
        if title is not None:
            g.fig.suptitle(title)
        if filename is not None:
            g.fig.savefig(filename, dpi=dpi)
            plt.close(g.fig)
        else:
            plt.show()

    def boxplot(self, **kwargs):
        """Boxplot of data.

        Notes: All keywords except further listed are passed to seaborn pairplot.
            Use all accessor valid columns or columns provided in `vars`

        Keyword Args:
            title (str): Title of the plot. Default None
            filename (str): If not none, plot is saved to file. Default None.
            dpi (int): DPI used for `savefig`. Default 150.
        """
        title = kwargs.pop("title", None)
        filename = kwargs.pop("filename", None)
        dpi = kwargs.pop("filename", 150)
        melt = {}

        if "vars" not in kwargs:
            melt["value_vars"] = None
            keep = []
        else:
            melt["value_vars"] = kwargs.pop("vars")
            keep = [c for c in melt["value_vars"] if c not in self._names]
        if "hue" in kwargs:
            hue = kwargs["hue"]
            melt["id_vars"] = [hue]
            keep.append(hue)

        m = pd.melt(self.df(keep=keep), **melt)
        g = sns.boxplot(m, x="variable", y="value", **kwargs)
        if title is not None:
            g.axes.title(title)
        if filename is not None:
            g.figure.savefig(filename, dpi=dpi)
            plt.close(g.figure)
        else:
            plt.show()

    def heatmap(self, **kwargs):
        """Correlation heatmap of data.

        Notes: All keywords except further listed are passed to seaborn pairplot.
            Use all accessor valid columns or columns provided in `vars`

        Keyword Args:
            title (str): Title of the plot. Default None
            filename (str): If not none, plot is saved to file. Default None.
            dpi (int): DPI used for `savefig`. Default 150.
        """
        title = kwargs.pop("title", None)
        filename = kwargs.pop("filename", None)
        dpi = kwargs.pop("filename", 150)

        if "vmin" not in kwargs:
            kwargs["vmin"] = -1
        if "vmax" not in kwargs:
            kwargs["vmax"] = 1
        if "cmap" not in kwargs:
            kwargs["cmap"] = "RdBu"
        if "annot" not in kwargs:
            kwargs["annot"] = True

        g = sns.heatmap(self._df.corr(), **kwargs)
        if title is not None:
            g.axes.title(title)
        if filename is not None:
            g.figure.savefig(filename, dpi=dpi)
            plt.close(g.fig)
        else:
            plt.show()


@pd.api.extensions.register_dataframe_accessor("oxides")
class OxidesAccessor(AccessorTemplate):
    """Use `.oxides` pandas dataframe accessor."""

    def _validate(self, obj):
        # verify there is a oxides column
        valid = []
        names = []
        props = []
        for col in obj.columns:
            try:
                f = formula(col)
                if (len(f.atoms) == 2) and (elements.name("oxygen") in f.atoms):
                    valid.append(True)
                    names.append(col)
                    props.append(oxideprops(f))
                else:
                    valid.append(False)
                    self._others.append(col)
            except (ValueError, pyparsing.exceptions.ParseException):
                valid.append(False)
                self._others.append(col)
        if not any(valid):
            raise MissingColumns("oxides")
        # sort names
        if config["sort_oxides"]:
            for ox in config["oxides_order"]:
                if ox in names:
                    ix = names.index(ox)
                    self._names.append(names.pop(ix))
                    self._names_props.append(props.pop(ix))
        self._names.extend(names)
        self._names_props.extend(props)

    def oxwt(self, **kwargs) -> pd.DataFrame:
        """Convert molar proportions to oxides weight percents.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with oxides weight percents

        """
        res = self._df.mul(self.props["mass"])
        return self._final(res, **kwargs)

    def cat_number(self, **kwargs) -> pd.DataFrame:
        """Calculate cations number.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with cations numbers

        """
        res = self.props["ncat"] * self._df.div(self.props["mass"])
        return self._final(res, **kwargs)

    def oxy_number(self, **kwargs) -> pd.DataFrame:
        """Calculate oxygens number.

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with oxygens numbers

        """
        res = self.props["noxy"] * self._df.div(self.props["mass"])
        return self._final(res, **kwargs)

    def omega(self, **kwargs) -> float:
        """Oxygen to positive atomic charges ratio

        Args:
            mineral (Mineral): noxy and ncat are taken from Mneral instance
            noxy (int): ideal number of oxygens. Default 1
            ncat (int): ideal number of cations. Default 1
        """
        if "mineral" in kwargs:
            ncat = kwargs.get("mineral").ncat
            noxy = kwargs.get("mineral").noxy
        else:
            ncat = kwargs.get("ncat", 1)
            noxy = kwargs.get("noxy", 1)

        return 2 * noxy / (self.charges(ncat).div(self.cnf(ncat), axis=0)).sum(axis=1)

    def onf(self, noxy: int) -> pd.Series:
        """Oxygen normalisation factor - ideal oxygens / sum of oxygens

        Keyword Args:
            noxy (int): ideal oxygens

        Returns:
            pandas.Series: oxygen normalisation factors

        """
        return noxy / self.oxy_number().sum(axis=1)

    def cnf(self, ncat: int) -> pd.Series:
        """Cation normalisation factor - ideal cations / sum of cations

        Args:
            ncat (int): ideal cations

        Returns:
            pandas.Series: cation normalisation factors

        """
        return ncat / self.cat_number().sum(axis=1)

    def cations(self, **kwargs) -> pd.DataFrame:
        """Cations calculated on the basis of oxygens or cations.

        Keyword Args:
            noxy (int): ideal number of oxygens. Default 1
            ncat (int): ideal number of cations. Default 1
            tocat (bool): when ``True`` normalized to ``ncat``,
                otherwise to ``noxy``. Default ``False``

        Returns:
            Dataframe with calculated cations

        """
        noxy = kwargs.get("noxy", 1)
        ncat = kwargs.get("ncat", 1)
        tocat = kwargs.get("tocat", False)
        if tocat:
            df = self.cat_number(dropna=False).multiply(self.cnf(ncat), axis=0)
            df.columns = [str(cat) for cat in self.props["cation"]]
            return self._final(df, **kwargs)
        else:
            df = self.cat_number(dropna=False).multiply(self.onf(noxy), axis=0)
            df.columns = [str(cat) for cat in self.props["cation"]]
            return self._final(df, **kwargs)

    def charges(self, ncat: int, **kwargs) -> pd.DataFrame:
        """Calculates charges based on number of cations.

        Args:
            ncat (int): number of cations

        Keyword Args:
            keep (list): list of additional columns to be included.Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with charges

        """
        charge = (
            self.cat_number(dropna=False).mul(self.cnf(ncat), axis=0)
            * self.props["charge"]
        )
        return self._final(charge, **kwargs)

    def charge_def(self, **kwargs) -> pd.Series:
        """Calculates charge deficiency based on number of oxygens and cations.

        Keyword Args:
            mineral (Mineral): noxy and ncat are taken from Mneral instance
            noxy (int): ideal number of oxygens. Default 1
            ncat (int): ideal number of cations. Default 1

        Returns:
            pandas.Series: Charge deficiency

        """
        if "mineral" in kwargs:
            ncat = kwargs.get("mineral").ncat
            noxy = kwargs.get("mineral").noxy
        else:
            ncat = kwargs.get("ncat", 1)
            noxy = kwargs.get("noxy", 1)
        charge = (
            self.cat_number(dropna=False).mul(self.cnf(ncat), axis=0)
            * self.props["charge"]
        )
        return 2 * noxy - charge.sum(axis=1)

    def apatite_correction(self, **kwargs) -> pd.DataFrame:
        """Apatite correction

        Note:
            All P2O5 is assumed to be apatite based and is removed from composition

                CaO mol% = CaO mol% - (10 / 3) * P2O5 mol%

        Keyword Args:
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Apatite corrected dataframe

        """
        if ("P2O5" in self._names) and ("CaO" in self._names):
            df = self._df.div(self.props["mass"])
            df = df.div(df.sum(axis=1), axis=0)
            df["CaO"] = (df["CaO"] - (10 / 3) * df["P2O5"]).clip(lower=0)
            df = df.mul(self.props["mass"], axis=1)
            df = df.drop(columns="P2O5")
            df = df.div(df.sum(axis=1), axis=0).mul(self._df.sum(axis=1), axis=0)
            return self._final(df, **kwargs)
        else:
            return self._final(self._df, **kwargs)

    def convert_Fe(self, **kwargs) -> pd.DataFrame:
        """Recalculate FeO to Fe2O3 or vice-versa.

        Note:
            When only FeO exists, all is recalculated to Fe2O3. When only Fe2O3
            exists, all is recalculated to FeO. When both exists, Fe2O3 is
            recalculated and added to FeO. Otherwise datatable is not changed.

        Keyword Args:
            to (str): to what iron oxide Fe should be converted. Default `"FeO"`
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with converted Fe oxide

        """
        to = kwargs.get("to", "FeO")
        if to == "FeO":
            if "Fe2O3" in self._names:
                Fe3to2 = 2 * formula("FeO").mass / formula("Fe2O3").mass
                res = self._df.copy()
                if "FeO" in self._names:
                    res["FeO"] += Fe3to2 * res["Fe2O3"]
                else:
                    res["FeO"] = Fe3to2 * res["Fe2O3"]
                res = res.drop(columns="Fe2O3")
                return self._final(res, **kwargs)
            else:
                return self._final(self._df, **kwargs)
        elif to == "Fe2O3":
            if "FeO" in self._names:
                Fe2to3 = formula("Fe2O3").mass / formula("FeO").mass / 2
                res = self._df.copy()
                if "Fe2O3" in self._names:
                    res["Fe2O3"] += Fe2to3 * res["FeO"]
                else:
                    res["Fe2O3"] = Fe2to3 * res["FeO"]
                res = res.drop(columns="FeO")
                return self._final(res, **kwargs)
            else:
                return self._final(self._df, **kwargs)
        else:
            print("Both FeO and Fe2O3 not in data. Nothing changed.")
            return self._final(self._df, **kwargs)

    def recalculate_Fe(self, **kwargs) -> pd.DataFrame:
        """Recalculate Fe based on charge balance.

        Note:
            Either both FeO and Fe2O3 are present or any of then, the composition
            is modified to fullfil charge balance for given cations and oxygens.

            Number of cations and oxygens could be provided by ncat and noxy args
            or by Mineral instance

        Keyword Args:
            mineral (Mineral): noxy and ncat are taken from Mneral instance
            noxy (int): ideal number of oxygens. Default 1
            ncat (int): ideal number of cations. Default 1
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with recalculated Fe

        """
        if "mineral" in kwargs:
            ncat = kwargs.get("mineral").ncat
            noxy = kwargs.get("mineral").noxy
        else:
            ncat = kwargs.get("ncat", 1)
            noxy = kwargs.get("noxy", 1)

        charge = self.cat_number(dropna=False).mul(self.cnf(ncat), axis=0)
        if ("Fe2O3" in self._names) & ("FeO" not in self._names):
            charge.loc[pd.isna(self._df["Fe2O3"]), "Fe2O3"] = 0
            chargedef = 2 * noxy - self.charges(ncat).sum(axis=1)
            toconv = chargedef
            charge["Fe2O3"] += toconv
            charge["FeO"] = -toconv
            ncats = self.props["ncat"]
            ncats["FeO"] = 1
            mws = self.props["mass"]
            mws["FeO"] = formula("FeO").mass
        elif "Fe2O3" in self._names:
            charge.loc[pd.isna(self._df["Fe2O3"]), "Fe2O3"] = 0
            chargedef = 2 * noxy - self.charges(ncat).sum(axis=1)
            toconv = chargedef.clip(lower=0, upper=charge["FeO"])
            charge["Fe2O3"] += toconv
            charge["FeO"] = charge["FeO"] - toconv
            ncats = self.props["ncat"]
            mws = self.props["mass"]
        elif "FeO" in self._names:
            chargedef = 2 * noxy - self.charges(ncat).sum(axis=1)
            charge["Fe2O3"] = chargedef.clip(lower=0, upper=charge["FeO"])
            charge["FeO"] = charge["FeO"] - charge["Fe2O3"]
            ncats = self.props["ncat"].copy()
            ncats["Fe2O3"] = 2
            mws = self.props["mass"].copy()
            mws["Fe2O3"] = formula("Fe2O3").mass
        else:
            print("No Fe in data. Nothing changed")
            return self._final(self._df, **kwargs)
        res = self._df.copy()
        ncharge = charge / ncat
        df = (
            ncharge.mul(mws)
            .mul(self.cat_number(dropna=False).sum(axis=1), axis="rows")
            .div(ncats)
        )
        res[df.columns] = df
        return self._final(res, **kwargs)

    def apfu(self, mineral: Mineral, **kwargs) -> pd.DataFrame:
        """Calculate a.p.f.u for given mineral

        Args:
            mineral: Mineral instance (see `petropandas.minerals`)

        Keyword Args:
            force (bool): when True, remaining cations are added to last site
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with apfu for given mineral

        """
        force = kwargs.get("force", False)
        if mineral.has_endmembers:
            if mineral.needsFe == "Fe2":
                dt = self.convert_Fe(**kwargs)
            elif mineral.needsFe == "Fe3":
                dt = self.recalculate_Fe(mineral=mineral, **kwargs)
            else:
                dt = self.df()
            cations = dt.oxides.cations(noxy=mineral.noxy, ncat=mineral.ncat, **kwargs)
            res = []
            for _, row in cations.ions.df().iterrows():
                res.append(mineral.apfu(row, force=force))
            return self._final(pd.DataFrame(res, index=self._obj.index), **kwargs)
        else:
            raise NoEndMembers(mineral)

    def check_stechiometry(self, mineral: Mineral, **kwargs) -> pd.Series:
        """Calculate average missfit of populated and ideal cations on sites for given mineral

        Args:
            mineral: Mineral instance (see `petropandas.minerals`)

        Keyword Args:
            force (bool): when True, remaining cations are added to last site

        Returns:
            Series with calculated misfit

        """
        force = kwargs.get("force", False)
        if mineral.has_endmembers:
            if mineral.needsFe == "Fe2":
                dt = self.convert_Fe(**kwargs)
            elif mineral.needsFe == "Fe3":
                dt = self.recalculate_Fe(mineral=mineral, **kwargs)
            else:
                dt = self.df()
            cations = dt.oxides.cations(noxy=mineral.noxy, ncat=mineral.ncat, **kwargs)
            res = []
            for _, row in cations.ions.df().iterrows():
                res.append(mineral.check_stechiometry(row, force=force))
            return pd.Series(res, index=self._obj.index, name="misfit")
        else:
            raise NoEndMembers(mineral)

    def endmembers(self, mineral: Mineral, **kwargs) -> pd.DataFrame:
        """Calculate endmembers proportions

        Args:
            mineral: Mineral instance (see `petropandas.minerals`)

        Keyword Args:
            tocat (bool): when ``True`` normalize to cations, otherwise to oxygens.
                Default ``False``
            force (bool): when True, remaining cations are added to last site
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with calculated endmembers

        """
        if "tocat" not in kwargs:
            kwargs["tocat"] = True
        force = kwargs.get("force", False)
        if mineral.has_endmembers:
            if mineral.needsFe == "Fe2":
                dt = self.convert_Fe(**kwargs)
            elif mineral.needsFe == "Fe3":
                dt = self.recalculate_Fe(mineral=mineral, **kwargs)
            else:
                dt = self.df()
            cations = dt.oxides.cations(noxy=mineral.noxy, ncat=mineral.ncat, **kwargs)
            res = []
            for _, row in cations.ions.df().iterrows():
                res.append(mineral.endmembers(row, force=force))
            return self._final(pd.DataFrame(res, index=self._obj.index), **kwargs)
        else:
            raise NoEndMembers(mineral)

    def TCbulk(self, **kwargs) -> None:
        """Print oxides formatted as THERMOCALC bulk script

        Note:
            The CaO is recalculate using apatite correction based on P205 if available.

        Keyword Args:
            H2O (float): wt% of water. When -1 the amount is calculated
                as 100 - Total. Default -1.
            oxygen (float): value to calculate moles of ferric iron.
                Moles FeO = FeOtot - 2O and moles Fe2O3 = O. Default 0.01
            system (str): axfile to be used. One of 'MnNCKFMASHTO', 'NCKFMASHTO',
                'KFMASH', 'NCKFMASHTOCr', 'NCKFMASTOCr'. Default 'MnNCKFMASHTO'

        """
        H2O = kwargs.get("H2O", -1)
        oxygen = kwargs.get("oxygen", 0.01)
        system = kwargs.get("system", "MnNCKFMASHTO")
        # fmt: off
        bulk = {
            "MnNCKFMASHTO": ["H2O", "SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "MnO", "O"],
            "NCKFMASHTO": ["H2O", "SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O"],
            "KFMASH": ["H2O", "SiO2", "Al2O3", "MgO", "FeO", "K2O"],
            "NCKFMASHTOCr": ["H2O", "SiO2", "Al2O3", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O", "Cr2O3"],
            "NCKFMASTOCr": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "TiO2", "O", "Cr2O3"],
        }
        # fmt: on
        assert system in bulk, "Not valid system"

        df = self.convert_Fe().oxides.apatite_correction()
        # Water
        if "H2O" in bulk[system]:
            if "H2O" not in df:
                if H2O == -1:
                    H2O = 100 - df.sum(axis=1)
                    H2O[H2O < 0] = 0
                else:
                    H2O = H2O * df.sum(axis=1) / (100 - H2O)
                df["H2O"] = H2O
        use = df.columns.intersection(bulk[system])
        df = df[use].oxides.molprop().oxides.scale(to=100 - oxygen)
        if "O" in bulk[system]:
            df["O"] = oxygen
        # add missing
        for lbl in bulk[system]:
            if lbl not in df:
                df[lbl] = 0.0
        print("bulk" + "".join([f"{lbl:>7}" for lbl in bulk[system]]))
        for ix, row in df[bulk[system]].iterrows():
            print("bulk" + "".join([f" {v:6.3f}" for v in row.values]) + f"  % {ix}")

    def Perplexbulk(self, **kwargs) -> None:
        """Print oxides formatted as PerpleX thermodynamic component list

        Note:
            The CaO is recalculate using apatite correction based on P205 if available.

        Keyword Args:
            H2O (float): wt% of water. When -1 the amount is calculated
                as 100 - Total. Default -1.
            oxygen (float): value to calculate moles of ferric iron.
                Moles FeO = FeOtot - 2O and moles Fe2O3 = O. Default 0.01
            system (str): system to be used. One of 'MnNCKFMASHTO', 'NCKFMASHTO',
                'KFMASH', 'NCKFMASHTOCr', 'NCKFMASTOCr'. Default 'MnNCKFMASHTO'

        """
        H2O = kwargs.get("H2O", -1)
        oxygen = kwargs.get("oxygen", 0.01)
        system = kwargs.get("system", "MnNCKFMASHTO")
        # fmt: off
        bulk = {
            "MnNCKFMASHTO": ["H2O", "SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "MnO", "O2"],
            "NCKFMASHTO": ["H2O", "SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O2"],
            "KFMASH": ["H2O", "SiO2", "Al2O3", "MgO", "FeO", "K2O"],
            "NCKFMASHTOCr": ["H2O", "SiO2", "Al2O3", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O2", "Cr2O3"],
            "NCKFMASTOCr": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "TiO2", "O2", "Cr2O3"],
        }
        # fmt: on
        assert system in bulk, "Not valid system"

        df = self.convert_Fe().oxides.apatite_correction()
        # Water
        if "H2O" in bulk[system]:
            if "H2O" not in df:
                if H2O == -1:
                    H2O = 100 - df.sum(axis=1)
                    H2O[H2O < 0] = 0
                else:
                    H2O = H2O * df.sum(axis=1) / (100 - H2O)
                df["H2O"] = H2O
        use = df.columns.intersection(bulk[system])
        df = df[use].oxides.molprop().oxides.scale(to=100 - oxygen)
        if "O2" in bulk[system]:
            df["O2"] = 2 * oxygen
        # add missing
        for lbl in bulk[system]:
            if lbl not in df:
                df[lbl] = 0.0
        print("begin thermodynamic component list")
        for ox, val in df[bulk[system]].iloc[0].items():
            print(f"{ox:6s}1 {val:8.5f}      0.00000      0.00000     molar amount")
        print("end thermodynamic component list")

    def MAGEMin(self, **kwargs) -> None:
        """Print oxides formatted as MAGEMin bulk file

        Note:
            The CaO is recalculate using apatite correction based on P205 if available.

        Keyword Args:
            H2O (float): wt% of water. When -1 the amount is calculated
                as 100 - Total. Default -1.
            oxygen (float): value to calculate moles of ferric iron.
                Moles FeO = FeOtot - 2O and moles Fe2O3 = O. Default 0.01
            db (str): MAGEMin database. 'mp' metapelite (White et al. 2014), 'mb' metabasite
                (Green et al. 2016), 'ig' igneous (Holland et al. 2018), 'um' ultramafic
                (Evans & Frost 2021), 'ume' ultramafic extended (Evans & Frost 2021 + pl, hb and aug
                from Green et al. 2016), 'mpe' Metapelite extended (White et al. 2014,
                Green et al. 2016, Evans & Frost 2021), 'mtl' mantle (Holland et al. 2013).
                Default is "mp"
            sys_in (str): system comp "wt" or "mol". Default is "mol"
            title (str): used as title. Default index
            comment (str): used as comment. Default 'petropandas'

        """
        H2O = kwargs.get("H2O", -1)
        oxygen = kwargs.get("oxygen", 0.01)
        db = kwargs.get("db", "mp")
        sys_in = kwargs.get("sys_in", "mol")
        title = kwargs.get("title", None)
        comment = kwargs.get("comments", "petropandas")
        # fmt: off
        bulk = {
            "ig": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O", "Cr2O3", "H2O"],
            "mp": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O", "MnO", "H2O"],
            "mb": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O", "H2O"],
            "um": ["SiO2", "Al2O3", "MgO", "FeO", "O", "H2O", "S"],
            "ume": ["SiO2", "Al2O3", "MgO", "FeO", "O", "H2O", "S", "CaO", "Na2O"],
            "mpe": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "K2O", "Na2O", "TiO2", "O", "MnO", "H2O", "CO2", "S"],
            "mtl": ["SiO2", "Al2O3", "CaO", "MgO", "FeO", "Na2O"],
        }
        # fmt: on
        assert db in bulk, "Not valid database"

        df = self.convert_Fe().oxides.apatite_correction()
        # Water
        if "H2O" in bulk[db]:
            if "H2O" not in df:
                if H2O == -1:
                    H2O = 100 - df.sum(axis=1)
                    H2O[H2O < 0] = 0
                else:
                    H2O = H2O * df.sum(axis=1) / (100 - H2O)
                df["H2O"] = H2O
        use = df.columns.intersection(bulk[db])
        if sys_in == "mol":
            df = df[use].oxides.molprop().oxides.scale(to=100 - oxygen)
        else:
            df = df[use].oxides.scale(to=100 - oxygen)
        if "O" in bulk[db]:
            df["O"] = oxygen
        # add missing
        for lbl in bulk[db]:
            if lbl not in df:
                df[lbl] = 0.0
        print("# HEADER")
        print("title; comments; db; sysUnit; oxide; frac; frac2")
        print("# BULK-ROCK COMPOSITION")
        for ix, row in df[bulk[db]].iterrows():
            oxides = ", ".join(row.keys())
            values = ", ".join([f"{val:.3f}" for val in row.values])
            if title is None:
                print(f"{ix};{comment};{db};{sys_in};[{oxides}];[{values}];")
            else:
                print(f"{title};{comment};{db};{sys_in};[{oxides}];[{values}];")


@pd.api.extensions.register_dataframe_accessor("ions")
class IonsAccessor(AccessorTemplate):
    """Use `.ions` pandas dataframe accessor."""

    def _validate(self, obj):
        # verify there is a ions column
        valid = []
        for col in obj.columns:
            try:
                f = formula(col)
                if (len(f.atoms) == 1) and (is_numeric_dtype(obj[col].dtype)):
                    if ision(next(iter(f.atoms.keys()))):
                        valid.append(True)
                        self._names.append(col)
                        self._names_props.append(elementprops(f))
                    else:
                        valid.append(False)
                        self._others.append(col)
                else:
                    valid.append(False)
                    self._others.append(col)
            except (ValueError, pyparsing.exceptions.ParseException):
                valid.append(False)
                self._others.append(col)
        if not any(valid):
            raise MissingColumns("ions")

    def wt(self, omega: float = 1) -> pd.DataFrame:
        """Oxides weight percents calculated from ions.

        Returns:
            Dataframe with calculated oxides weight percents

        """
        df = pd.DataFrame()
        for col, prop in zip(self._names, self._names_props):
            ion = next(iter(formula(col).atoms.keys()))
            m, n = ionox[ion.charge]
            ox = formula(f"{ion.element.symbol}{m}O{n}")
            df[str(ox)] = self._obj[col] * ox.mass / (omega * ox.structure[0][0])
        return df


@pd.api.extensions.register_dataframe_accessor("elements")
class ElementsAccessor(AccessorTemplate):
    """Use `.elements` pandas dataframe accessor."""

    def _validate(self, obj):
        # verify there is an element column
        valid = []
        for col in obj.columns:
            try:
                f = formula(col)
                if (len(f.atoms) == 1) and (is_numeric_dtype(obj[col].dtype)):
                    if iselement(next(iter(f.atoms.keys()))):
                        valid.append(True)
                        self._names.append(col)
                        self._names_props.append(elementprops(f))
                    else:
                        valid.append(False)
                        self._others.append(col)
                else:
                    valid.append(False)
                    self._others.append(col)
            except (ValueError, pyparsing.exceptions.ParseException):
                valid.append(False)
                self._others.append(col)
        if not any(valid):
            raise MissingColumns("elements")


@pd.api.extensions.register_dataframe_accessor("ree")
class REEAccessor(AccessorTemplate):
    """Use `.ree` pandas dataframe accessor."""

    def _validate(self, obj):
        # verify there is a REE column
        valid = []
        for col in obj.columns:
            if col in REE:
                valid.append(True)
                self._names.append(col)
                self._names_props.append(elementprops(formula(col)))
            else:
                valid.append(False)
                self._others.append(col)
        if not any(valid):
            raise MissingColumns("REE")

    def normalize(self, **kwargs) -> pd.DataFrame:
        """Normalize REE by reservoir.

        Note:
            Predefined reservoirs are imported from
            [GERM Reservoir Database](https://earthref.org/GERMRD/reservoirs/). You can
            check all available reservoirs in `config["reservoirs"]`.

        Keyword Args:
            reservoir (str): Name of reservoir. Deafult "CI Chondrites"
            reference (str): Reference. Default "McDonough & Sun 1995"
            source (str): Original source. Deafult same as reference.
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True

        Returns:
            Dataframe with normalized REE composition
        """
        reservoir = kwargs.get("reservoir", "CI Chondrites")
        reference = kwargs.get("reference", "McDonough & Sun 1995")
        source = kwargs.get("source", reference)
        nrm = pd.Series(config["reservoirs"][reservoir][reference][source])
        res = self._df / nrm
        res = res[self._names]
        res["Eu/Eu*"] = res["Eu"] / np.sqrt(res["Sm"] * res["Gd"])
        res["Gd/Yb"] = res["Gd"] / res["Yb"]
        return self._final(res, **kwargs)

    def plot(self, **kwargs):
        """Spiderplot of REE data.

        Note:
            List of REE used for plot could be set in `config["ree_plot"]`

        Keyword Args:
            grouped (bool): When True aggegated data with confidence interval is drawn.
                Default False
            boxplot (bool): When True, boxplot for each REE is drawn. Default False
            boxplot_props (dict): Additional arguments passed to `sns.boxplot`. Default
                `{"color": "grey"}`.
            hue (str or None): Name of columns used for colors.
            palette (string, list, dict, or matplotlib.colors.Colormap): Method for
                choosing the colors to use when mapping the hue semantic.
            legend ("auto", "brief", "full", or False): How to draw the legend.
            title (str): Title of the plot. Default None
            filename (str): If not none, plot is saved to file. Default None.
            dpi (int): DPI used for `savefig`. Default 150.
            keep (list): list of additional columns to be included. Default [].
            dropna (bool): whether to drop columns with NA only. Default True
        """
        fig, ax = plt.subplots()
        ndf = self.normalize(**kwargs)
        ax.set(yscale="log")
        ree = ndf.melt(
            id_vars=kwargs.get("keep", []),
            var_name="ree",
            ignore_index=False,
        )
        # select only REE for plotting
        ree = ree.loc[ree["ree"].isin(config["ree_plot"])]
        if kwargs.get("grouped", False):
            sns.lineplot(
                data=ree,
                x="ree",
                y="value",
                hue=kwargs.get("hue", None),
                palette=kwargs.get("palette", None),
                errorbar="ci",
                legend=kwargs.get("legend", "brief"),
                ax=ax,
            )
        else:
            sns.lineplot(
                x="ree",
                y="value",
                data=ree,
                hue=kwargs.get("hue", None),
                palette=kwargs.get("palette", None),
                units=ree.index,
                estimator=None,
                legend=kwargs.get("legend", "brief"),
                ax=ax,
            )
        if kwargs.get("boxplot", False):
            sns.boxplot(
                data=ree,
                x="ree",
                y="value",
                flierprops={"ms": 3},
                ax=ax,
                **kwargs.get("boxplot_props", {"color": "grey"}),
            )
        ax.set_title(kwargs.get("title", ""))
        ax.set_xlabel("")
        if "filename" in kwargs:
            fig.tight_layout()
            fig.savefig(
                kwargs.get("filename", "ree_plot.pdf"), dpi=kwargs.get("dpi", 150)
            )
            plt.close(fig)
        else:
            plt.show()


if __name__ == "__main__":  # pragma: no cover
    pass



import importlib.resources

import pandas as pd

res = importlib.resources.files("petropandas").joinpath("data")

src = res.joinpath("oxides", "avgpelite.csv")
with importlib.resources.as_file(src) as f:
    avgpelite = pd.read_csv(f)

src = res.joinpath("oxides", "bulk.csv")
with importlib.resources.as_file(src) as f:
    bulk = pd.read_csv(f)

src = res.joinpath("oxides", "grt_profile.csv")
with importlib.resources.as_file(src) as f:
    grt_profile = pd.read_csv(f)

src = res.joinpath("oxides", "minerals.csv")
with importlib.resources.as_file(src) as f:
    minerals = pd.read_csv(f)

src = res.joinpath("oxides", "pyroxenes.csv")
with importlib.resources.as_file(src) as f:
    pyroxenes = pd.read_csv(f)

src = res.joinpath("mnz", "sbdata.csv")
with importlib.resources.as_file(src) as f:
    mnz_sb = pd.read_csv(f)



import numpy as np
import pandas as pd


def read_actlabs(src, **kwargs):
    if "skiprows" not in kwargs:
        kwargs["skiprows"] = 2
    df = pd.read_excel(src, **kwargs).rename(columns={"Fe2O3(T)": "Fe2O3"})
    units = df.iloc[0]
    limits = df.iloc[1]
    method = df.iloc[2]
    df = df.rename(columns={"Analyte Symbol": "Sample"})[3:].set_index("Sample")
    # replace detection limits
    for col in df:
        ix = df[col].astype(str).str.startswith("< ")
        if any(ix):
            df.loc[ix, col] = np.nan

    df = df.astype(float)
    return df, units, limits, method


def read_bureau_veritas(src=""):
    df = pd.read_excel(src, skiprows=9)
    method = df.iloc[0][2:]
    cols = df.iloc[1]
    cols[:2] = ["Sample", "Type"]
    units = df.iloc[2][2:]
    limits = df.iloc[3][2:]
    selection = df.iloc[:, 1] == "Rock Pulp"

    dt = df.iloc[:, 2:][selection]
    # replace detection limits
    for col in dt:
        dt[col][dt[col].astype(str).str.startswith("<") is True] = 0

    dt = dt.astype(float).copy()

    res = pd.concat([df[selection].iloc[:, :2], dt], axis=1)
    res.columns = cols.str.strip()
    return res, units, limits, method



"""Client for postresql petrodb database API

This module provide easy CRUD operations on petrodb database

Usage examples:

    from petropandas import pd
    from petropandas.database import PetroDB

    db = PetroDB('http://127.0.0.1:8000', 'user', 'password')
    project = db.projects(name="MyProject")
    sample = project.samples(name="DB250")
    df = sample.spots.df(mineral="Grt")

"""

from functools import cached_property

import pandas as pd
import requests


def zero_negative_nan(x):
    if isinstance(x, (int, float, complex)) and not isinstance(x, bool):
        if x > 0:
            return x
        else:
            return float("nan")
    else:
        return x


class PetroAPI:
    def __init__(self, api_url, username, password):
        response = requests.post(
            f"{api_url}/token",
            data={"username": username, "password": password},
        )
        if response.ok:
            self.__api_url = api_url
            self.__username = username
            self.__password = password
            self.logged = True
        else:
            self.logged = False

    def __authorize(self):
        if self.logged:
            response = requests.post(
                f"{self.__api_url}/token",
                data={"username": self.__username, "password": self.__password},
            )
            if response.ok:
                token = response.json()
                return {"Authorization": f"Bearer {token.get('access_token')}"}
            else:
                raise ValueError("Wrong url or credentials")
        else:
            raise ConnectionError("Not logged in")

    def get(self, path):
        headers = self.__authorize()
        return requests.get(f"{self.__api_url}/api{path}", headers=headers)

    def post(self, path, data):
        headers = self.__authorize()
        return requests.post(f"{self.__api_url}/api{path}", json=data, headers=headers)

    def put(self, path, data):
        headers = self.__authorize()
        return requests.put(f"{self.__api_url}/api{path}", json=data, headers=headers)

    def delete(self, path):
        headers = self.__authorize()
        return requests.delete(f"{self.__api_url}/api{path}", headers=headers)


class PetroDB:
    """Petro database instance

    High-level access to online Petro database

    """

    def __init__(self, api_url, username, password):
        self.__db = PetroAPI(api_url, username, password)

    def __repr__(self):
        return f"PetroDB {'OK' if self.__db.logged else 'Not logged'}"

    def projects(self, name: str | None = None):
        """Get project from database

        Args:
            name (str, optional): search for project by name

        Returns:
            Project instance or list of all projects if the name is not provided.

        Raises:
            ValueError: If project(s) was not found.

        """
        if name is not None:
            response = self.__db.get(f"/search/project/{name}")
            if response.ok:
                return PetroDBProject(self.__db, project=response.json())
            else:
                raise ValueError(response.json()["detail"])
        else:
            response = self.__db.get("/projects/")
            if response.ok:
                return [PetroDBProject(self.__db, project=p) for p in response.json()]
            else:
                raise ValueError(response.json()["detail"])

    def create_project(self, name: str, description: str = ""):
        """Create project in database

        Args:
            name (str): name of the project
            description (str, optional): decription of the project. Default ``.

        Returns:
            Created project instance.

        Raises:
            ValueError: If project was not created.

        """
        data = {"name": name, "description": description}
        response = self.__db.post("/project/", data)
        if response.ok:
            return PetroDBProject(self.__db, project=response.json())
        else:
            raise ValueError(response.json()["detail"])


class PetroDBProject:
    """Petro DB project instance

    Attributes:
        data (dict): project attributes

    """

    def __init__(self, db, project):
        self.__db = db
        self.__project_id = project.pop("id")
        self.data = project

    def __repr__(self):
        return f"{self.name}"

    @property
    def name(self):
        """str: Name of the project."""
        return self.data["name"]

    @name.setter
    def name(self, name: str):
        self.data["name"] = name

    @property
    def description(self):
        """str: Description of the project."""
        return self.data["description"]

    @description.setter
    def description(self, desc: str):
        self.data["description"] = desc

    def samples(self, name: str | None = None):
        """Get sample from database

        Args:
            name (str, optional): search for sample by name

        Returns:
            Sample instance or list of all samples if the name is not provided.

        Raises:
            ValueError: If sample(s) was not found.

        """
        if name is not None:
            response = self.__db.get(f"/search/sample/{self.__project_id}/{name}")
            if response.ok:
                return PetroDBSample(
                    self.__db, self.__project_id, sample=response.json()
                )
            else:
                raise ValueError(response.json()["detail"])
        else:
            response = self.__db.get(f"/samples/{self.__project_id}")
            if response.ok:
                return [
                    PetroDBSample(self.__db, self.__project_id, sample=s)
                    for s in response.json()
                ]
            else:
                raise ValueError(response.json()["detail"])

    def create_sample(self, name: str, description: str = ""):
        """Create project in database

        Args:
            name (str): name of the sample
            description (str, optional): decription of the sample. Default ``.

        Returns:
            Created sample instance.

        Raises:
            ValueError: If sample was not created.

        """
        data = {"name": name, "description": description}
        response = self.__db.post(f"/sample/{self.__project_id}", data)
        if response.ok:
            return PetroDBSample(self.__db, self.__project_id, sample=response.json())
        else:
            raise ValueError(response.json()["detail"])

    def delete(self):
        """Delete project from database. Use with caution!"""
        response = self.__db.delete(f"/project/{self.__project_id}")
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update project in database according to the data attribute."""
        response = self.__db.put(f"/project/{self.__project_id}", self.data)
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    @property
    def spots(self):
        """Spots accessor"""
        res = []
        names = []
        for sample in self.samples():
            try:
                res += list(sample.spots.records.values())
                names += sample.spots.sample
            except ValueError:
                pass
        return PetroDBSpotRecords(res, sample=names)

    @property
    def areas(self):
        """Areas accessor"""
        res = []
        names = []
        for sample in self.samples():
            try:
                res += list(sample.areas.records.values())
                names += sample.areas.sample
            except ValueError:
                pass
        return PetroDBAreaRecords(res, sample=names)

    def mineral_data(self, mineral: str):
        """Return spots and profile spots of given mineral dataframe"""
        res = []
        samples = self.samples()
        for sample in samples:
            spots = sample.spots.df(mineral=mineral, sample_name=True)
            if not spots.empty:
                res.append(spots)
            profiles = sample.profiles(mineral=mineral)
            for profile in profiles:
                res.append(profile.spots.df(sample_name=True))
        return pd.concat(res)

    def add_user(self, username: str):
        """Add access to the project to user

        Args:
            username (str): username

        """
        response = self.__db.put(
            f"/project/{self.__project_id}/adduser", {"username": username}
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def remove_user(self, username: str):
        """Remove access to the project for user

        Args:
            username (str): username

        """
        response = self.__db.put(
            f"/project/{self.__project_id}/removeuser", {"username": username}
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])


class PetroDBSample:
    """Petro DB sample instance

    Attributes:
        data (dict): project attributes

    """

    def __init__(self, db, project_id, sample):
        self.__db = db
        self.__project_id = project_id
        self.__sample_id = sample.pop("id")
        self.data = sample

    def __repr__(self):
        return f"{self.name}"

    @property
    def name(self):
        """str: Name of the sample."""
        return self.data["name"]

    @name.setter
    def name(self, name: str):
        self.data["name"] = name

    @property
    def description(self):
        """str: Description of the sample."""
        return self.data["description"]

    @description.setter
    def description(self, desc: str):
        self.data["description"] = desc

    @cached_property
    def spots(self):
        """Spots accessor"""
        response = self.__db.get(f"/spots/{self.__project_id}/{self.__sample_id}")
        if response.ok:
            return PetroDBSpotRecords(
                response.json(), sample=len(response.json()) * [self.name]
            )
        else:
            raise ValueError(response.json()["detail"])

    def spot(self, spot_id: int):
        """Get spot from database

        Args:
            spot_id (int): id of the spot

        Returns:
            Spot instance.

        Raises:
            ValueError: If spot was not found.

        """
        response = self.__db.get(
            f"/spot/{self.__project_id}/{self.__sample_id}/{spot_id}"
        )
        if response.ok:
            return PetroDBSpot(
                self.__db, self.__project_id, self.__sample_id, spot=response.json()
            )
        else:
            raise ValueError(response.json()["detail"])

    def create_spot(
        self,
        label: str,
        mineral: str,
        values: dict,
    ):
        """Create spot in database

        Args:
            label (str): label of the spot
            mineral (str): Name of mineral. Kretz abbreviations recommended
            values (dict): Data values

        Returns:
            Created spot instance.

        Raises:
            ValueError: If spot was not created.

        """
        data = {"label": label, "mineral": mineral, "values": values}
        response = self.__db.post(f"/spot/{self.__project_id}/{self.__sample_id}", data)
        if response.ok:
            return PetroDBSpot(
                self.__db, self.__project_id, self.__sample_id, spot=response.json()
            )
        else:
            raise ValueError(response.json()["detail"])

    def create_spots(
        self,
        df: pd.DataFrame,
        label_col: str | None = None,
        mineral_col: str | None = None,
    ):
        """Create spots in database from pandas dataframe

        Args:
            df (pandas.DataFrame): data values
            label_col (str, optional): Name of column to be used as label. If not
                provided, dataframe index is used
            mineral_col (str, optional): Name of column to be used as mineral. If not
                provided, mineral will be empty

        Returns:
            Created spots

        Raises:
            ValueError: If spots were not created.

        """
        df = df.copy()
        if label_col is None:
            labels = pd.Series(df.index.astype(str), index=df.index)
        else:
            labels = df[label_col].str.strip()
            df.drop(label_col, axis=1, inplace=True)
        if mineral_col is None:
            minerals = pd.Series("", index=df.index)
        else:
            minerals = df[mineral_col].str.strip()
            df.drop(mineral_col, axis=1, inplace=True)
        spots = []
        for label, mineral, (ix, row) in zip(labels, minerals, df.iterrows()):
            spots.append(
                {
                    "label": label,
                    "mineral": mineral,
                    "values": row.apply(zero_negative_nan).dropna().to_dict(),
                }
            )
        response = self.__db.post(
            f"/spots/{self.__project_id}/{self.__sample_id}", spots
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    @cached_property
    def areas(self):
        """Areas accessor"""
        response = self.__db.get(f"/areas/{self.__project_id}/{self.__sample_id}")
        if response.ok:
            return PetroDBAreaRecords(
                response.json(), sample=len(response.json()) * [self.name]
            )
        else:
            raise ValueError(response.json()["detail"])

    def area(self, area_id: int):
        """Get area from database

        Args:
            area_id (int): id of the spot

        Returns:
            Area instance.

        Raises:
            ValueError: If area was not found.

        """
        response = self.__db.get(
            f"/area/{self.__project_id}/{self.__sample_id}/{area_id}"
        )
        if response.ok:
            return PetroDBArea(
                self.__db, self.__project_id, self.__sample_id, area=response.json()
            )
        else:
            raise ValueError(response.json()["detail"])

    def create_area(self, label: str, values: dict):
        """Create area in database

        Args:
            label (str): label of the area
            values (dict): Data values

        Returns:
            Created area instance.

        Raises:
            ValueError: If area was not created.

        """
        data = {"label": label, "values": values}
        response = self.__db.post(f"/area/{self.__project_id}/{self.__sample_id}", data)
        if response.ok:
            return PetroDBArea(
                self.__db, self.__project_id, self.__sample_id, area=response.json()
            )
        else:
            raise ValueError(response.json()["detail"])

    def create_areas(
        self,
        df: pd.DataFrame,
        label_col: str | None = None,
    ):
        """Create areas in database from pandas dataframe

        Args:
            df (pandas.DataFrame): data values
            label_col (str, optional): Name of column to be used as label. If not
                provided, dataframe index is used

        Returns:
            Created areas

        Raises:
            ValueError: If areas were not created.

        """
        df = df.copy()
        if label_col is None:
            labels = pd.Series(df.index.astype(str), index=df.index)
        else:
            labels = df[label_col].str.strip()
            df.drop(label_col, axis=1, inplace=True)
        areas = []
        for label, (ix, row) in zip(labels, df.iterrows()):
            areas.append(
                {
                    "label": label,
                    "values": row.apply(zero_negative_nan).dropna().to_dict(),
                }
            )
        response = self.__db.post(
            f"/areas/{self.__project_id}/{self.__sample_id}", areas
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def profiles(self, label: str | None = None, mineral: str | None = None):
        """Get profile from database

        Args:
            label (str, optional): search for sample by name

        Returns:
            Profile instance or list of all profiles if the label is not provided.

        Raises:
            ValueError: If profile(s) was not found.

        """
        if label is not None:
            response = self.__db.get(
                f"/search/profile/{self.__project_id}/{self.__sample_id}/{label}"
            )
            if response.ok:
                data = response.json()
                if mineral is not None:
                    if data["mineral"] == mineral:
                        return PetroDBProfile(
                            self.__db,
                            self.__project_id,
                            self.__sample_id,
                            self.name,
                            profile=response.json(),
                        )
                    else:
                        raise ValueError(
                            f"Profile with {label=} and {mineral=} not found."
                        )
                return PetroDBProfile(
                    self.__db,
                    self.__project_id,
                    self.__sample_id,
                    self.name,
                    profile=response.json(),
                )
            else:
                raise ValueError(response.json()["detail"])
        else:
            response = self.__db.get(
                f"/profiles/{self.__project_id}/{self.__sample_id}"
            )
            if response.ok:
                if mineral is not None:
                    return [
                        PetroDBProfile(
                            self.__db,
                            self.__project_id,
                            self.__sample_id,
                            self.name,
                            profile=p,
                        )
                        for p in response.json()
                        if p["mineral"] == mineral
                    ]
                else:
                    return [
                        PetroDBProfile(
                            self.__db,
                            self.__project_id,
                            self.__sample_id,
                            self.name,
                            profile=p,
                        )
                        for p in response.json()
                    ]
            else:
                raise ValueError(response.json()["detail"])

    def create_profile(self, label: str, mineral: str):
        """Create profile in database

        Args:
            label (str): label of the profile
            mineral (str): Name of mineral. Kretz abbreviations recommended

        Returns:
            Created profile instance.

        Raises:
            ValueError: If profile was not created.

        """
        data = {"label": label, "mineral": mineral}
        response = self.__db.post(
            f"/profile/{self.__project_id}/{self.__sample_id}", data
        )
        if response.ok:
            return PetroDBProfile(
                self.__db,
                self.__project_id,
                self.__sample_id,
                self.name,
                profile=response.json(),
            )
        else:
            raise ValueError(response.json()["detail"])

    def delete(self):
        """Delete sample from database. Use with caution!"""
        response = self.__db.delete(f"/sample/{self.__project_id}/{self.__sample_id}")
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update sample in database according to the data attribute."""
        response = self.__db.put(
            f"/sample/{self.__project_id}/{self.__sample_id}", self.data
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def reset(self):
        """Reset cached properties spots and areas to access updated data"""
        if "spots" in self.__dict__:
            self.__dict__["spots"]
        if "areas" in self.__dict__:
            self.__dict__["areas"]

    @property
    def profilespots(self):
        res = []
        names = []
        for profile in self.profiles():
            try:
                res += list(profile.spots.records.values())
                names += profile.spots.sample
            except ValueError:
                pass
        return PetroDBProfilespotRecords(res, sample=names)


class PetroDBSpot:
    """Petro DB spot instance

    Attributes:
        data (dict): project attributes

    """

    def __init__(self, db, project_id, sample_id, spot):
        self.__db = db
        self.__project_id = project_id
        self.__sample_id = sample_id
        self.__spot_id = spot.pop("id")
        self.data = spot

    def __repr__(self):
        return f"{self.label} ({self.mineral})"

    @property
    def label(self):
        """str: Label of the spot."""
        return self.data["label"]

    @label.setter
    def label(self, lbl: str):
        self.data["label"] = lbl

    @property
    def mineral(self):
        """str: Mineral of the spot."""
        return self.data["mineral"]

    @mineral.setter
    def mineral(self, m: str):
        self.data["mineral"] = m

    def delete(self):
        """Delete spot from database. Use with caution!"""
        response = self.__db.delete(
            f"/spot/{self.__project_id}/{self.__sample_id}/{self.__spot_id}"
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update spot in database according to the data attribute."""
        response = self.__db.put(
            f"/spot/{self.__project_id}/{self.__sample_id}/{self.__spot_id}",
            self.data,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])


class PetroDBArea:
    """Petro DB area instance

    Attributes:
        data (dict): project attributes

    """

    def __init__(self, db, project_id, sample_id, area):
        self.__db = db
        self.__project_id = project_id
        self.__sample_id = sample_id
        self.__area_id = area.pop("id")
        self.data = area

    def __repr__(self):
        return f"{self.label}"

    @property
    def label(self):
        """str: Label of the area."""
        return self.data["label"]

    @label.setter
    def label(self, lbl: str):
        self.data["label"] = lbl

    def delete(self):
        """Delete area from database. Use with caution!"""
        response = self.__db.delete(
            f"/area/{self.__project_id}/{self.__sample_id}/{self.__area_id}"
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update area in database according to the data attribute."""
        response = self.__db.put(
            f"/area/{self.__project_id}/{self.__sample_id}/{self.__area_id}",
            self.data,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])


class PetroDBProfile:
    """Petro DB sample instance

    Attributes:
        data (dict): project attributes
        samplename (str): name of the sample

    """

    def __init__(self, db, project_id, sample_id, samplename, profile):
        self.__db = db
        self.__project_id = project_id
        self.__sample_id = sample_id
        self.samplename = samplename
        self.__profile_id = profile.pop("id")
        self.data = profile

    def __repr__(self):
        return f"{self.label} ({self.mineral})"

    @property
    def label(self):
        """str: Label of the profile."""
        return self.data["label"]

    @label.setter
    def label(self, lbl: str):
        self.data["label"] = lbl

    @property
    def mineral(self):
        """str: Mineral of the profile."""
        return self.data["mineral"]

    @mineral.setter
    def mineral(self, m: str):
        self.data["mineral"] = m

    @cached_property
    def spots(self):
        """Profile spots accessor"""
        response = self.__db.get(
            f"/profilespots/{self.__project_id}/{self.__sample_id}/{self.__profile_id}"
        )
        if response.ok:
            recs = response.json()
            for rec in recs:
                rec["label"] = self.label
            return PetroDBProfilespotRecords(recs, sample=len(recs) * [self.samplename])
        else:
            raise ValueError(response.json()["detail"])

    def spot(self, spot_id: int):
        """Get profile spot from database

        Args:
            spot_id (int): id of the profile spot

        Returns:
            Profile spot instance.

        Raises:
            ValueError: If profile spot was not found.

        """
        response = self.__db.get(
            f"/profilespot/{self.__project_id}/{self.__sample_id}/{self.__profile_id}/{spot_id}"
        )
        if response.ok:
            return PetroDBProfileSpot(spot=response.json(), **self._kwargs)
        else:
            raise ValueError(response.json()["detail"])

    def create_spot(self, index: int, values: dict):
        """Create profile spot in database

        Args:
            index (int): used to define order of spots on profile
            values (dict): Data values

        Returns:
            Created profile spot instance.

        Raises:
            ValueError: If profile spot was not created.

        """
        data = {"index": index, "values": values}
        response = self.__db.post(
            f"/profilespot/{self.__project_id}/{self.__sample_id}/{self.__profile_id}",
            data,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def create_spots(self, df: pd.DataFrame):
        """Create profile spots in database from pandas dataframe

        Args:
            df (pandas.DataFrame): data values, index must be numeric and is used to define order

        Returns:
            Created profile spots

        Raises:
            ValueError: If profile spots were not created.

        """
        df = df.copy()
        profilespots = []
        for index, row in df.iterrows():
            profilespots.append(
                {
                    "index": index,
                    "values": row.apply(zero_negative_nan).dropna().to_dict(),
                }
            )
        response = self.__db.post(
            f"/profilespots/{self.__project_id}/{self.__sample_id}/{self.__profile_id}",
            profilespots,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def delete(self):
        """Delete profile from database. Use with caution!"""
        response = self.__db.delete(
            f"/profile/{self.__project_id}/{self.__sample_id}/{self.__profile_id}"
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update profile in database according to the data attribute."""
        response = self.__db.put(
            f"/profile/{self.__project_id}/{self.__sample_id}/{self.__profile_id}",
            self.data,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def reset(self):
        """Reset cached property spots to access updated data"""
        if "spots" in self.__dict__:
            self.__dict__["spots"]


class PetroDBProfileSpot:
    """Petro DB profile spot instance

    Attributes:
        data (dict): project attributes

    """

    def __init__(self, db, project_id, sample_id, profile_id, spot):
        self.__db = db
        self.__project_id = project_id
        self.__sample_id = sample_id
        self.__profile_id = profile_id
        self.__profilespot_id = spot.pop("id")
        self.data = spot

    def __repr__(self):
        return f"Spot {self.index}"

    @property
    def index(self):
        """int: Index of the profile spot."""
        return self.data["index"]

    @index.setter
    def index(self, idx: int):
        self.data["index"] = idx

    def delete(self):
        """Delete profile spot from database. Use with caution!"""
        response = self.__db.delete(
            f"/profilespot/{self.__project_id}/{self.__sample_id}/{self.__profile_id}/{self.__profilespot_id}"
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def update(self):
        """Update profile spot in database according to the data attribute."""
        response = self.__db.put(
            f"/profilespot/{self.__project_id}/{self.__sample_id}/{self.__profile_id}/{self.__profilespot_id}",
            self.data,
        )
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])


class PetroDBRecords:
    """Petro DB records accessor"""

    def __init__(self, recs: list, sample: list):
        self.records = {rec["id"]: rec for rec in recs}
        self.sample = sample
        self.cols = []

    def __repr__(self):
        return f"{len(self.records)} spots"

    def df(self, **kwargs):
        """Get records as pandas dataframe

        Note: Keyword arguments `col=val` are used to select records with given value.
            `col` must be on of the available columns in attribute cols.

        Attributes:
            cols (list): list of attributes for selection

        """
        res = pd.DataFrame(
            [row["values"] for row in self.records.values()], index=self.records.keys()
        )
        res["sample"] = self.sample
        for col in self.cols:
            res[col] = [row[col] for row in self.records.values()]
        for col, val in kwargs.items():
            if col in self.cols:
                res = res[res[col] == val]
        return res.sort_index().copy()


class PetroDBSpotRecords(PetroDBRecords):
    def __init__(self, recs: list, sample: list):
        super().__init__(recs, sample)
        self.cols = ["label", "mineral"]


class PetroDBAreaRecords(PetroDBRecords):
    def __init__(self, recs: list, sample: list):
        super().__init__(recs, sample)
        self.cols = ["label"]


class PetroDBProfilespotRecords(PetroDBRecords):
    def __init__(self, recs: list, sample: list):
        super().__init__(recs, sample)
        self.cols = ["label"]


class PetroDBAdmin:
    """Admin client for postresql petrodb database API."""

    def __init__(self, api_url, username, password):
        self.__db = PetroAPI(api_url, username, password)

    # ---------- USERS

    def users(self, name: str | None = None):
        response = self.__db.get("/users/")
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])

    def create_user(self, username: str, password: str, email: str):
        """Create user in database

        Args:
            username (str): username
            password (str): password
            email (str): email

        """
        data = {"username": username, "password": password, "email": email}
        response = self.__db.post("/user/", data)
        if response.ok:
            return response.json()
        else:
            raise ValueError(response.json()["detail"])



import pytest

from petropandas import mindb, pd  # noqa: F401
from petropandas.data import grt_profile, minerals  # noqa: F401

grt = mindb.Garnet_Fe2()


@pytest.fixture
def simple():
    return pd.DataFrame(
        [[1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6]],
        columns=["SiO2", "TiO2", "La", "Gd", "Lu", "F"],
    )


def test_OxidesAccessor(simple):
    assert all(simple.oxides._df.columns == ["SiO2", "TiO2"])


def test_ElementsAccessor(simple):
    assert all(simple.elements._df.columns == ["La", "Gd", "Lu", "F"])


def test_REEAccessor(simple):
    assert all(simple.ree._df.columns == ["La", "Gd", "Lu"])


def test_endmembers():
    assert pytest.approx(grt_profile.oxides.endmembers(grt).sum().sum()) == len(
        grt_profile
    )


def test_endmembers_keep():
    assert grt_profile.oxides.endmembers(grt, keep=["Label"]).shape == (99, 5)


def test_check_stechiometry():
    assert (
        pytest.approx(grt_profile.oxides.check_stechiometry(grt).sum())
        == 0.0008756712880324041
    )


def test_search():
    assert len(minerals.petro.search("pl-", on="Comment")) == 3
